# Harmonize

- [Overview](#overview)
- [What does it do?](#what-does-it-do)
- [Should I use it?](#should-i-use-it)
- [Quick Start](#quick-start)
- [Fine-Tuning your data](#fine-tuning-your-data)
- [Upgrading V2 -> V3](#upgrading-from-v2-to-v3)
- [Features](#features)
  - [Module's Main Settings](#modules-main-settings)
  - [Entity Settings](#entity-settings)
  - [Data Alteration](#data-alteration)
  - [Twig Functions & Filters](#twig-functions--filters)
  - [Entity Processing Rules](#entity-processing-rules)
  - [Styles](#styles)
  - [Caching](#caching)
  - [Visualizer](#visualizer)
- [DIY Zone](#diy-zone)
  - [The Harmonize Service](#the-harmonize-service)

## Overview
Harmonize streamlines preprocessing of entities in Drupal before template
rendering. It is a tool for developers and was originally made to facilitate
collaboration between backend and frontend teams.

## What does it do?
Harmonize makes it much easier to access field values in Drupal templates. It
does this by preprocessing entity content and adding a new variable called
`harmony` to your twig templates. This variable contains a simple, cleanly
formatted array of data to work with.

Drupal already exposes the values of your content to your templates, but it is
sometimes difficult or tedious to access said data. This is especially true when
trying to access values in nested entities.

Take this example. Here we're fetching the first 3 articles entered in a
`field_related_articles` field of a `node` template. After fetching these
articles, we access their thumbnails and want to print their path, alt and title
values.

### Standard

```html
<div>
  {% for articleFieldValue in node.field_related_articles[0:3] %}
    <img
      src="{{ file_url(articleFieldValue.entity.field_thumbnail.entity.field_media_image.entity.fileuri) }}"
      alt="{{ articleFieldValue.entity.field_thumbnail.entity.field_media_image.alt }}"
      title="{{ articleFieldValue.entity.field_thumbnail.entity.field_media_image.title }}"
    />
  {% endfor %}
</div>
```

### With Harmonize (With fine-tuning)

```html
<div>
  {% for article in harmony.field_related_articles[0:3] %}
    <img
      src="{{ article.field_thumbnail.src }}"
      alt="{{ article.field_thumbnail.alt }}"
      title="{{ article.field_thumbnail.title }}"
    />
  {% endfor %}
</div>
```

If this already makes you a little happier, then good! Keep on reading to see
how you can achieve this using Harmonize.

## Should I use it?

Harmonize is a module with tons of features developed over time. At its core, it
greatly facilitates preprocessing for templates.

If you want to say goodbye to 2000 line `.theme` files, then this might be the
module for you.

Another great use case for this module is to help save time for devs that are more focused
on templating and styling. Essentially, with Harmonize, you
can bring non-drupal front-end devs onto Drupal projects with ease. It can be quite
a learning curve (and headache inducing experience) for front-end devs to try to
access certain entity values. As seen in the example above, it's much easier to
play with data that is more readable.

## Quick Start

_The examples below are done in a test website called `dreamland` and use a
theme called `dreamland_theme`._

In this section, we'll work our way towards what Harmonize can accomplish with
no additional code.

To see how this module works out of the box, you can simply enable it and set up
simple content configurations.

Once this is done, edit any entity type. In the example below, I am
editing a **Content Type** called **Page**.

![Content Type Settings](./doc/screenshots/Screenshot_1.png)

Check the first checkbox shown in the screenshot to enable the module to
preprocess data and assign it to your templates.

Optionally, you can check the second checkbox and then visit your content type's
display settings to confirm which fields are preprocessed.

By default, all fields are processed.

![Content Type Display Settings](./doc/screenshots/Screenshot_2.png)

Finally, you can now inspect the contents of the `harmony` variable either
inside a `node.html.twig` or in a `hook_preprocess_node` function.

In this example, we'll opt for the latter to avoid any memory errors caused by
Twig's default dumper. (If you're on Drupal 9.5+, you can safely use `dump()`
in Templates).

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node(&$variables) {
  dump($variables['harmony']);
}
```

Or in Twig (Safe to do in Drupal 9.5.x+)

```html
{{ dump(harmony) }}
```

The data should be shown to you when visiting your page. In the case of the
website I'm using for this documentation, the data looks like this.

![Data Example 1](./doc/screenshots/Screenshot_3.png)

Now say that on my **Page** templates, I want to display a section that shows
the 3 first articles in the `field_related_articles` field. I can now do
something like this.

```html
<div>
  {% for article in harmony.field_related_articles[0:3] %}
    <img
      src="{{ article|get('field_thumbnail').field_media_image.attributes.src }}"
      alt="{{ article|get('field_thumbnail').field_media_image.attributes.alt }}"
      title="{{ article|get('field_thumbnail').field_media_image.attributes.title }}"
    />
  {% endfor %}
</div>
```

And that's it! Now this is the out-of-the-box configuration, but there are
ways to optimize the data even further, as seen [above](#what-does-it-do).

For instance, you may notice that in the data example shown here, the entity
reference fields return an unfamiliar class.

These classes are defined in the module and are called **Harmonizers**. They
comprise the backbone of the module and are used to extract data from entities
and other Drupal objects.

This is why in our twig, we make use of the `|get` filter. That is a custom
filter provided by Harmonize to fetch data from Harmonizers.

Every Harmonizer also has a `::harmonize()` function that can be called to get
the data of the entity it is intended to process.

If we were to fine-tune the example above with the following hook:

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node(&$variables) {
  // Process nested entities with Harmonizers.
  // We do it for field_related_articles.
  $variables['harmony']['field_related_articles'] = array_map(function ($harmonizer) {
    return $harmonizer->harmonize();
  }, $variables['harmony']['field_related_articles']);

  // And for field_featured_articles.
  $variables['harmony']['field_featured_articles'] = array_map(function ($harmonizer) {
    return $harmonizer->harmonize();
  }, $variables['harmony']['field_featured_articles']);
}
```

Then the data would look more like:

![Data Example 2](./doc/screenshots/Screenshot_6.png)

And your twig template could be refactored to be like this:

```html
<div>
    {% for article in harmony.field_related_articles[0:3] %}
        <img
            src="{{ article.field_thumbnail.field_media_image.attributes.src }}"
            alt="{{ article.field_thumbnail.field_media_image.attributes.alt }}"
            title="{{ article.field_thumbnail.field_media_image.attributes.title }}"
        />
    {% endfor %}
</div>
```

Looks better already right?

Now let's take it up a notch.

## Fine-tuning your data

This section will outline more optimization techniques this module can help you
achieve. We will stick with the same example as above.

As a refresher, let's go through the current state of our code so far.

We have the following hook in our `.theme` file.

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node(&$variables) {
  // Process nested entities with Harmonizers.
  $variables['harmony']['field_related_articles'] = array_map(function ($harmonizer) {
    return $harmonizer->harmonize();
  }, $variables['harmony']['field_related_articles']);

  $variables['harmony']['field_featured_articles'] = array_map(function ($harmonizer) {
    return $harmonizer->harmonize();
  }, $variables['harmony']['field_featured_articles']);
}
```

And our twig template looks like this:

```html
<div>
    {% for article in harmony.field_related_articles[0:3] %}
        <img
            src="{{ article.field_thumbnail.field_media_image.attributes.src }}"
            alt="{{ article.field_thumbnail.field_media_image.attributes.alt }}"
            title="{{ article.field_thumbnail.field_media_image.attributes.title }}"
        />
    {% endfor %}
</div>
```

Let's optimize this even further.

### Optimizing Media Processing

Media references to images (and other types) usually nest the information we
want very deep. A good example of what I'm getting at is the following:

`node.field_related_articles.0.entity.field_thumbnail.entity.field_media_image.entity.fileuri`

This is how we would traditionally access the URI of the thumbnail in the first
related article of our node. The entity nesting goes deep, which makes accessing
the URI quite a hassle.

With the configurations we have so far, it's already a little better. In our
Twig, we can do the following:

`article.field_thumbnail.field_media_image.attributes.src`

But what if we want to eliminate that call to `field_media_image`? We know that
for all `image` medias across our site, the actual image file information will
be in the `field_media_image` field, so let's try to eliminate that extra step.

There are many ways to do this. The best Harmonize feature to use in this case
would be a [**Harmony File**](#harmony-files). These files allow you to alter
processed data for given entity types, field types and more. In this case, we
want to alter the data processed when Harmonize encounters Image medias.

To achieve this, I can create a `entity.media.image.harmony.php` file in my
`dreamland_theme` theme. The full path to the file and its contents would be:

`web/themes/custom/dreamland_theme/harmony/entity.media.image.harmony.php`
```php
<?php

/** @file... */

// Alter our data.
$harmony = [
  'type' => 'image',
  'uri' => $consignment['field_media_image']['uri'] ?? '',
  'src' => $consignment['field_media_image']['attributes']['src'] ?? '',
  'attributes' => $consignment['field_media_image']['attributes'] ?? [],
  'alt' => $consignment['field_media_image']['attributes']['alt'] ?? '',
  'title' => $consignment['field_media_image']['attributes']['title'] ?? '',
];
```

In this file, we are altering what Harmonize returns when **Media Entities** are
processed.

Now with this, our final data will now look more like this:

![Data Example 3](./doc/screenshots/Screenshot_12.png)

Which permits us to now do this in our templates!

```html
<div>
    {% for article in harmony.field_related_articles[0:3] %}
        <img
            src="{{ article.field_thumbnail.src }}"
            alt="{{ article.field_thumbnail.alt }}"
            title="{{ article.field_thumbnail.title }}"
        />
    {% endfor %}
</div>
```

And with (hopefully) easy to understand steps, we've reached some clean data!

This optimization showcases [Harmony Files](#harmony-files), one of the two ways
Harmonize offers to fine-tune data. You can read more about them below.

### Harmonize Baseline

You can opt to enable the **Harmonize Baseline** module. This module does some
out of the box data optimizations for certain entity types. It is a quite simple
module that makes use of Harmonize's existing features to optimize the data.

For the media optimizations just addressed earlier, this module does them out of
the box for you.

If you do this, then the data output you see will already be optimized like how
we saw in the example at the top of this guide.
**I highly recommend doing this** as it just saves you time in the long run.

### Do we need the hook?

Let's go back to our hook now.

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node(&$variables) {
  // Process nested entities with Harmonizers.
  $variables['harmony']['field_related_articles'] = array_map(function ($harmonizer) {
    return $harmonizer->harmonize();
  }, $variables['harmony']['field_related_articles']);

  $variables['harmony']['field_featured_articles'] = array_map(function ($harmonizer) {
    return $harmonizer->harmonize();
  }, $variables['harmony']['field_featured_articles']);
}
```

Now to re-iterate, this hook is done to return data from **Harmonizers**. You
will usually encounter **Harmonizers** when dealing with `entity_reference`
fields.

The reason why Harmonize returns Harmonizers instead of the data from the nested
entity directly is simply to avoid risks of infinite loops. This is discussed
further in the documentation [below](#automatically-processed-entities).

This being said, it can get tedious to write hooks to do this every time. There
are two other ways to get around this in the module.

#### Using the `|harmonize` or `|get()` filters (Twig Solution)

In Twig, you have access to two functions that can help you get data from a
**Harmonizer**. See the two examples below that will achieve the same result as
we have above, but without the hook being implemented at all.

```html
<div>
    {% for article in harmony.field_related_articles[0:3] %}
        {% set article = harmonize(article) %}
        <img
            src="{{ article|harmonize.field_thumbnail.src }}"
            alt="{{ article|harmonize.field_thumbnail.alt }}"
            title="{{ article|harmonize.field_thumbnail.title }}"
        />
    {% endfor %}
</div>
```

OR

```html
<div>
    {% for article in harmony.field_related_articles[0:3] %}
        <img
            src="{{ article|get('field_thumbnail').src }}"
            alt="{{ article|get('field_thumbnail').alt }}"
            title="{{ article|get('field_thumbnail').title }}"
        />
    {% endfor %}
</div>
```

With this, we can remove the hook from our code and still access the data properly.

#### Configuration Solution

Alternatively, you can set up processing rules for fields in entities.

At `/admin/structure/harmonize/entity-rules`, you can tell Harmonize to
automatically process referenced entities in the `field_related_articles` field
for example.

Once that is done, you can go back to having Twig like the following:

```html
<div>
    {% for article in harmony.field_related_articles[0:3] %}
        <img
            src="{{ article.field_thumbnail.src }}"
            alt="{{ article.field_thumbnail.alt }}"
            title="{{ article.field_thumbnail.title }}"
        />
    {% endfor %}
</div>
```

This solution will not necessarily lead to infinite loops and won't be dangerous
since Harmonize will only process one level. Any further nested entities that
may exist in Articles will return Harmonizers as normal.

### Want to optimize it further?

Well, if you're still hungry for more, then I'm pleased to say that there are
many more features available to you. Refer to any of the sections below to find
out what more you can do to preprocess your data!

* [Refiners](#refiners) - An alternative to [Harmony Files](#harmony-files)
* [Styles](#styles) - Harmonize's built-in field filtering solution.
* [Caching](#caching) - How Harmonize caches data.
* [More on Entity Processing Rules](#entity-processing-rules)
* [Visualizer](#visualizer) - Testing and troubleshooting data preprocessing.
* More coming soon! Promise!

## Upgrading from V2 to V3

So you're working on a website that has V2 installed, and you want to upgrade to
V3...

The first thing to note is that there are many differences in V3. One of the
main things is that nested preprocessing of entities in entity reference fields
no longer happens automatically. This is because V2 had a lot of risks with this
behaviour. Infinity loops were much more prevalent.

With this major change, you will have to refactor most of your existing code
pertaining to harmonize.

When `entity_reference` fields are processed, a **Harmonizer** is returned
instead of the expected data of the referenced entities. To reach the nested
data, you simply have to call `->harmonize()` on the Harmonizer object. Keep
this in mind when refactoring your code.

### Upgrade Steps

Follow these steps to assure the least amount of headache when upgrading.

* Assure site configurations in your codebase are up-to-date: `drush cex -y`.
* Update the module via composer: `composer update drupal/harmonize`.
* Run database updates: `drush updb -y`.
* **Important** - Export migrated configurations.
* **3.1** - Rename all `.harmony.php` files to `.harmony`.
* Now you should be good to go...To start refactoring!

### Changelog

Below are other things to consider when upgrading to V3.

* Nested processing of referenced entities is no longer automatic. Adjust code accordingly.
* Harmonize Admin links have been moved under **Structure** instead of **Configuration**.
* Harmonizer Signatures have been fine-tuned. When targeting entities in a specific style, you must adjust your harmony file names and Refiner targets.
  * `entity.node.article--card` becomes `entity.node.article.card`.
* Data configurations have been deprecated. They have been replaced with [Entity Processing Rules](#entity-processing-rules).
* Refiners should be good as they are thankfully.
* Theme Harmony configuration YAML files have been adjusted.
  * `THEMENAME.harmony.yml` is no longer loaded. The module will now load two separate files.
  * `THEMENAME.harmony.styles.yml` for Styles.
  * `THEMENAME.harmony.rules.yml` for Entity Rules.
  * See the `harmonize_examples` module for examples of the file structures. Examples will also be available in this guide soon.
* Theme `.harmony.php` files have been adjusted. Harmonize will now look for `.harmony` files instead.
  * You will have to change all `.harmony.php` file extensions to simply `.harmony`.
* An `$object` variable is now available in `.harmony` files. This variable contains the original object being preprocessed.
* [Styles](#styles) can now be created in the admin interface.
* A [Visualizer](#visualizer) has been added to allow entity processing tests.
* New [Twig](#twig-functions--filters) functionality has been added.
  * This may help with refactoring in Twig templates.
* [Caching](#caching) has been optimized and is no longer experimental. Enable it. Trust me.
* [Entity Settings](#entity-settings) have been enabled along with view mode integration.
  * Coming from V2, this may not be needed. But it may be something to consider during refactoring.
* [Main Module Settings](#modules-main-settings) have been adjusted.
  * Moved Automatic Preprocessing setting to a textarea. You now specify which entity types are automatically processed in entity reference fields.
    * **WARNING**: This setting can lead to infinite loops in some cases. Be mindful of the entity types you add here.
  * Removed recursive depth settings. No longer needed.
  * Added new setting for data alteration object loading. See [this section](#data-alteration) for more details.
    * If you're coming from V2 and you have a lot of data alteration objects, you may want to set this to [**SCAN**](#scan-elevator).

## Features

### Module's Main Settings

The main settings of the module can be configured at
`/admin/structure/harmonize/settings`.

![Main Settings](./doc/screenshots/Screenshot_4.png)
![Main Settings 2](./doc/screenshots/Screenshot_5.png)

Below is a description of each setting.

#### Harmony File Discovery

This enables [Harmony File Discovery](#harmony-files). Refer to the
[section](#harmony-files) specific to this below.

#### Loading Algorithm

This configures the loading algorithm. Don't mind this for now as this will be
covered in the [Data Alteration](#data-alteration) section.

#### Automatically Processed Entities

This allows you to configure which entities are automatically processed. If you
remember the [Quick Start Example](#quick-start) above, you saw that entity
reference fields values return a **Harmonizer**. If the entity type being
processed is in this setting, it will return the processed data directly
instead.

Here is an example of the data from the Quick Start above if `node.article` is
added in this setting.

![Data Example 2](./doc/screenshots/Screenshot_6.png)

Instead of returning a **Harmonizer**, we now see the data of the referenced
entities in the entity reference fields.

The module normally returns **Harmonizers** to avoid surprise infinity loops in
entity reference fields. Say an article in my example had an entity reference
field of its own pointing towards a page, there would be risk of infinity loops
if entities were automatically processed.

With that in mind, be wary of the entities you add to this setting.

The recommended and safer way to set auto-preprocessing on nested entities is
with [Entity Processing Rules](#entity-processing-rules). Refer to that section
for more information on that.

Alternatively, you can always accomplish this manually by doing something
similar to this:

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node(&$variables) {
  // Manually process nested entities.
  $variables['harmony']['field_related_articles'] = array_map(function ($articleHarmonizer) {
    return $articleHarmonizer->harmonize();
  }, $variables['harmony']['field_related_articles']);
}
```

All **Harmonizers** can be used to access nested data with the `->harmonize()`
function.

#### Allowed Admin Routes

Harmonize normally does not execute in admin routes as it is defined to act
before the templating layer of the frontend. With that being said, there are
cases where you do want it to execute in an admin route. Layout builder paths
are a good example of this since to preview blocks while editing content, Drupal
needs to render the entities for you to see.

Note that the layout builder routes in question are currently already hardcoded,
so you don't need to add them yourself. Chances are this setting won't need to
be touched for a while.

### Entity Settings

As seen in the [Quick Start](#quick-start), you have access to entity specific
settings. These will usually be seen when editing an entity type.

![Content Type Settings](./doc/screenshots/Screenshot_1.png)

#### Populate Theme Variables with Harmonized Data

This quite simply automatically adds the `harmony` key to the `$variables` array
of your entity's template. It is the equivalent of writing the following hook:

_This example assumes you are doing this for **Nodes** of type `page`._

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node__page(&$variables) {
  // Manually process nested entities.
  $variables['harmony'] = \Drupal::service('harmonize')->harmonize($variables['node'], ['view_mode' => $variables['view_mode']]);
}
```

#### Enable display filtering

Grants access to filtering which fields are preprocessed per display type. See
the screenshot below.

![Content Type Display Settings](./doc/screenshots/Screenshot_2.png)

New fields will be included by default. You simply uncheck the checkbox on any
field you want to omit. Disabled fields with the checkbox checked will still be
preprocessed.

### Data Alteration

Now we're getting to the good stuff!

Everything we've seen so far is what can be done without any additional code.
Harmonize offers ways to fine-tune the data before it is sent to templates.

Now the easiest way of course would be through your traditional Drupal Hooks.

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node(&$variables) {
  // Get the data from Harmonize.
  $harmony = &$variables['harmony'];

  // Do data alteration stuff...
  $harmony['related_article_count'] = count($harmony['field_related_articles']);
}
```

This is completely fine to do, but Harmonize offers other ways to do this
without writing hooks. The two main ways to do this are:

* Harmony Files
* Refiners

Henceforth, these will be referred to as **Alteration Objects**.

#### Harmony Files

**Harmony Files** are the main way to alter data before it is sent to templates.
You need the [**Harmony File Discovery**](#harmony-file-discovery) setting to be
enabled for this feature to be available to you.

With the feature enabled, Harmonize will look for files in your active theme.
The file crawling is similar to how Drupal's templating system works.

**e.g.** For a `node` of type `page`, you would create a `node--page.html.twig`
in the `templates` directory of your theme.

**cont.** For the same `node` of type `page`, you would create a
`entity.node.page.harmony` file in the `harmony` directory of your theme.

The contents of the `entity.node.page.harmony` file would be as follows:

```php
<?php

use Drupal\harmonize\Service\Harmonics;

/**
 * @file
 * Override & customize data harmonizations for a given Drupal object or entity.
 *
 * @harmonizerType entity
 * @entityType node
 * @bundle page
 *
 *
 * This is a custom file to handle the reformatting of data before it reaches
 * Drupal templates.
 *
 * Available variables:
 * @var mixed $consignment
 *    The data array (modify in place).
 *    This is comparable to the '$variables' array in preprocess functions.
 * @var mixed $harmony
 *    The final data that CAN be returned to the Twig template.
 *    Usage of this variable is optional, and it is empty by default.
 *    If this is left empty, then the data in $consignment will be returned.
 * @var mixed $object
 *    The original object being harmonized. Sometimes you want to access the
 *    original Drupal object if Harmonize does not provide certain information.
 * @var \Drupal\harmonize\Service\Harmonics $harmonics
 *    The Harmonics service which contains helper functions.
 *
 * You have the choice to either modify the $consignment array which contains
 * the default data, or assign new data to the empty $harmony variable. If
 * $harmony is empty by the end of this file, $consignment will be sent to
 * templates. Otherwise, $harmony will be sent to templates.
 */

// The same alterations we did earlier, but in a harmony file!
$consignment['related_article_count'] = count($consignment['field_related_articles']);
```

The `$consignment` variable contains all the data so far. The `$harmony`
variable is the data that will be returned to templates. If `$harmony` is empty
by the end of the file, `$consignment` will be returned. So you have two options
in Harmony Files:

* Modify `$consignment` and never set `$harmony`. This is good for small touch-ups of the data.
* Set `$harmony` variable with new data. This is good if you have to **rearrange** data or make bigger changes.

In our example above, we are simply adding new data, so we modify
`$consignment`.

And voila! In your `node--page.html.twig` template, you should now have access
to `harmony.related_article_count`.

Note that Harmony Files run before your traditional hooks, so you can always do
a combination of both.

If the Harmony File above exists, then:

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node(&$variables) {
  // Since the harmony file declared above exists, this should evaluate to true.
  if (isset($harmony['related_article_count'])) {
    // Yay! It exists!
  }
}
```

As you may have guessed, the file crawling works similar to template
suggestions. So `entity.node.harmony` would work too, except it would apply
to all nodes instead of just `page` typed nodes. You can also specify a view
mode, like `entity.node.page.default` or `entity.node.page.teaser`.

#### Refiners

Refiners are alteration objects that work very similarly to harmony files. They
are quite simply your OOP alternative to them, allowing for Dependency Injection
and more.

Refiners use the
[Plugin API](https://www.drupal.org/docs/drupal-apis/plugin-api).

If you haven't read the [Harmony Files](#harmony-files) section, I suggest you
do so now to get a good comparison!

First, you would create a plugin in a module. For the sake of this example, I am
creating a module called `dreamland_refiners`.

Then, I can create a plugin at
`dreamland_refiners/src/Plugin/harmonize/Refiner/PageNodeRefiner.php`.

The contents would be the following:

```php
<?php

namespace Drupal\dreamland_refiners\Plugin\harmonize\Refiner;

use Drupal\harmonize\PluginManager\Refinery\RefinerBase;

/**
 * Provide plugin to alter harmonizer data for an entity.
 *
 * @noinspection
 *   AnnotationMissingUseInspection
 *
 * @Refiner(
 *   id = "dreamland_refiners.entity.node.page.refiner",
 *   target = "entity.node.page"
 * )
 *
 * @package Drupal\harmonize\Plugin\Harmonizer\Refiner
 */
final class PageNodeRefiner extends RefinerBase {

  /**
   * Add personal tweaks to harmonized data in this function.
   *
   * {@inheritdoc}
   */
  public function refine(&$consignment, &$harmony, $object): void {
    // Similar alterations to what we did earlier, but in a refiner!
    $consignment['featured_article_count'] = count($consignment['field_featured_articles']);
  }

}
```

Refiners work quite exactly like Harmony Files. Anything found inside the
`$harmony` variable at the end of the `refine()` function's execution will be
sent to templates. If `$harmony` isn't set, `$consignment` will be taken
instead.

The refiner here knows what to target due to the annotation definition. You can
see that the target is set to `entity.node.page`.

This works just like Harmony files. You could set it to `entity.node` to target
all nodes, or `entity.node.page.default` for the `default` display type.

#### Alteration Object Flow

So now we should have a good understanding of data alteration objects. With
this, we can finally talk about what happens when all of them exist.

Let's say we have everything from the examples above, so:

* A Refiner at `dreamland_refiners/src/Plugin/harmonize/Refiner/PageNodeRefiner.php` with the same content as above.
* A Harmony File at `dreamland_theme/harmony/entity.node.page.harmony` with the same content as above.

The final data would be the following:

![Data Example 3](./doc/screenshots/Screenshot_7.png)

But you're probably wondering what goes first? The answer to your question is
right here.

`Harmonize Preprocesses Entity -> Refiners Alter Data -> Harmony Files Alter Data -> Data is sent to 'harmony' key in $variables -> Data is sent to twig templates`

This flow promotes the following best practices:

* Use Harmonize for easy base preprocessing of entities.
* Create Refiners for big data alterations that require advanced Drupal service injection, database queries and more.
* Create Harmony Files OR preprocess functions in your theme for final touch-ups of data before it reaches templates.
* Use `harmony` in your templates to access the preprocessed data.

#### Alteration Object Loading Algorithm Explained

Remember that [setting](#loading-algorithm) from above? Now we can finally talk
about how it works.

Harmonize has two ways to load alteration objects. To showcase these two
methodologies, let's say we have the following:

* A harmony file called `entity.node.harmony.php`.
* A harmony file called `entity.node.page.harmony`.
* We're viewing a `node` of type `page`.

Refiners work in the same way, so any explanations below apply to them too.

##### FCFS (First Come First Serve)

**FCFS** will **only load the first alteration object found**. It attempts to
load **the most relevant alteration object** for a given template.

So in the case above, only `entity.node.page.harmony` will be loaded,
because you're loading a `node` of type `page`.

Other harmony files will be ignored once one is loaded.

This loading philosophy is akin to core loading philosophies for templates for
example.

This is also the **recommended** setting to use.

##### SCAN (Elevator)

**SCAN** will **load multiple alteration objects**. It loads **all relevant
alteration objects** for a given template.

So in the case above, both `entity.node.page.harmony` and
`entity.node.harmony` will be loaded. It will process them in order of how
specific your alteration object is, from least to most specific. Here,
`entity.node.harmony` will be loaded first, and then
`entity.node.page.harmony` will be loaded after.

This can be useful to avoid code duplication in some cases, but it is not the
recommended setting to use. **SCAN** essentially exists as a backwards
compatibility option if you're migrating from V2. When starting new projects,
definitely opt to use FCFS.

### Twig Functions & Filters

Harmonize currently offers one twig function and two twig filters.

#### harmonize()

The `harmonize()` function can be used on an entity in a template to get
harmonized data.

For example in `node.html.twig`, we usually have access to the `node` variable,
which contains the node being viewed.

If you don't have pre-process hooks and want to harmonize the data, you can do
the following.

```html
{% set harmony = harmonize(node) %}
```

This is something that you can do if you don't want Harmonize to do things
automatically. You can also use this function to get harmonized data from
entities in Entity Reference fields.

#### |get('FIELD_NAME')

The `|get('FIELD_NAME')` filter can be used from a **Harmonizer** to get a field
value from a Harmonizer.

This is the same example as the one at the top of this guide. Here, we use it to
get harmonized data on the referenced articles.

```html
<div>
    {% for article in harmony.field_related_articles[0:3] %}
        <img
            src="{{ article|get('field_thumbnail').field_media_image.attributes.src }}"
            alt="{{ article|get('field_thumbnail').field_media_image.attributes.alt }}"
            title="{{ article|get('field_thumbnail').field_media_image.attributes.title }}"
        />
    {% endfor %}
</div>
```

#### |harmonize

The `|harmonize` filter can be used from a **Harmonizer** to get harmonized
data.

```html
<div>
    {% for article in harmony.field_related_articles[0:3] %}
        <img
            src="{{ article|harmonize.field_media_image.attributes.src }}"
            alt="{{ article|harmonize.field_media_image.attributes.alt }}"
            title="{{ article|harmonize.field_media_image.attributes.title }}"
        />
    {% endfor %}
</div>
```

### Styles

**More info soon!**

### Entity Processing Rules

Entity Processing Rules can be managed at
`admin/structure/harmonize/entity-rules`.

The settings here allow you to specify certain behaviors when Harmonize
processes fields. You can:

* Exclude a field from all processing.
* Disable caching on a field.
* Automatically process referenced entities (Entity Reference Fields Only)
* Set a Harmonizer Style (Entity Reference Fields Only)

This page is great for setting global processing rules for entities. For
example, if there is a field that you never want Harmonize to process, this is
where you would set it.

**e.g.** Excluding all Layout Builder fields from processing in nodes.

You can set up rules for Entities and Entity Bundles.

### Caching

With caching enabled, Harmonize will cache preprocessing results. This is
**heavily recommended** to keep on simply for performance reasons.

Harmonize results will share cache tags with the object it processes.

You can configure the cache at `admin/structure/harmonize/cache`.

#### Entity Caching

When Harmonizing an Entity, caching works a little differently. Instead of
caching the full data result, caching is managed on a **per field basis**.

All field value caching is tied to the entity they've been entered in, as they
should be.

Additionally, **entity reference fields are not directly cached**. This is
because the referenced entities' field values are cached individually
cached instead.

### Visualizer

The Visualizer is a powerful tool that can be used to test entity preprocessing.

You can configure the visualizer at `admin/structure/harmonize/visualize`.

**More information soon.**

## DIY Zone

Now the **DIY** (Do It Yourself) zone is reserved for advanced backend
developers that want to do things their way.

So far, everything that was shown can be accomplished even by developers that
are more specialized in Front-End and templating. The stuff covered in this
section is intended for back-end developers that seek to dive deeper into how
Harmonize works and seek to potentially develop solutions of their own based off
the inner-workings of the module.

### The Harmonize Service

The **Harmonize** service is the central point of the module. In other words,
any object being preprocessed by the module is passed through the
`::harmonize()` method available in this service.

To see what I mean, take this example below.

```php
/**
 * Implements hook_preprocess_HOOK() for node.html.twig.
 */
function dreamland_theme_preprocess_node(&$variables) {
  $variables['harmony'] = Drupal::service('harmonize')->harmonize($variables['node']);
}
```

If you were to enable the module without configuring any of the settings, you
could use this hook to preprocess a node before template rendering. This would
achieve pretty much the same result as if you had enabled your node type to be
preprocessed via [Entity Settings](#entity-settings).

The `::harmonize()` function accepts nearly any entity. I say nearly because
I personally haven't tested all entity types the world has to offer. However,
the module doesn't do anything crazy depending on the entity type being
processed. It simply loops through the fields and gets their values in a clean
format.
