<?php

namespace Drupal\harmonize\Service\FormTuner;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;

/**
 * Provides a form tuner for Entity Type Manage Display forms.
 *
 * @package Drupal\harmonize\Service\FormTuner
 */
final class EntityTypeViewDisplayFormTuner extends FormTuner {

  /**
   * ID of this form tuner.
   */
  public const ID = 'harmonize.form_tuner.entity_type_view_display';

  /**
   * {@inheritDoc}
   *
   * @var \Drupal\Core\Entity\Display\EntityViewDisplayInterface $formEntity
   */
  public function alter(array &$form, FormStateInterface $formState, EntityInterface $formEntity) : void {
    // Get entity type ID and bundle.
    $entityTypeId = $formEntity->getTargetEntityTypeId();
    $bundle = $formEntity->getTargetBundle();

    // Our goal here is to modify the table to include additional options.
    // First, we check if we have Harmonize settings.
    $settings = FormTuner::loadEntitySettings($entityTypeId, $bundle);

    // If settings aren't specified, or we don't have display filtering enabled,
    // we can stop here.
    if (empty($settings) || !$settings->getAllowDisplayFiltering()) {
      return;
    }

    // Attach library.
    $form['#attached']['library'][] = 'harmonize/entity_type_view_display_form';

    // Get the display mode.
    $displayMode = $formEntity->getMode();

    // Get display settings configurations.
    $displaySettings = $settings->getDisplaySettings();

    // Get the table.
    $table = &$form['fields'];

    // Add header for preprocess inclusion.
    $table['#header'][] = $this->t('Harmonize');

    // Add new header for harmonize settings.
    $table['#header'][] = [
      'data'    => '',
      'colspan' => 2,
    ];

    // Loop into our rows and add the checkbox.
    foreach ($table as $fieldName => $rowRenderArray) {
      // We don't act on render array keys.
      if (str_starts_with($fieldName, '#')) {
        continue;
      }

      // If the field doesn't start with 'field_', we skip.
      // We take 'body' though. We like 'body'.
      if ($fieldName !== 'body' && !str_starts_with($fieldName, 'field_')) {
        continue;
      }

      // Add harmonize checkbox.
      $table[$fieldName]['harmonize'] = [
        '#type' => 'checkbox',
        '#default_value' => $displaySettings[$displayMode][$fieldName] ?? TRUE,
        '#attributes' => ['class' => ['harmonize-entity-settings-harmonize-checkbox']],
        '#wrapper_attributes' => ['class' => ['harmonize-entity-settings-harmonize-checkbox-wrapper']],
      ];

      // Add harmonize settings.
//      if ($formState->get('harmonize_settings_edit') == $field_name) {
//        // We are currently editing this field's harmonize settings. Display the
//        // settings form and submit buttons.
//        $field_row['harmonize_plugin']['settings_edit_form'] = [];
//
//        if ($plugin) {
//          // Generate the settings form and allow other modules to alter it.
//          $settings_form = $plugin->settingsForm($form, $form_state);
//          $third_party_settings_form = $this->thirdPartySettingsForm($plugin, $field_definition, $form, $form_state);
//
//          if ($settings_form || $third_party_settings_form) {
//            $field_row['plugin']['#cell_attributes'] = ['colspan' => 3];
//            $field_row['plugin']['settings_edit_form'] = [
//              '#type' => 'container',
//              '#attributes' => ['class' => ['field-plugin-settings-edit-form']],
//              '#parents' => ['fields', $field_name, 'settings_edit_form'],
//              'label' => [
//                '#markup' => $this->t('Plugin settings'),
//              ],
//              'settings' => $settings_form,
//              'third_party_settings' => $third_party_settings_form,
//              'actions' => [
//                '#type' => 'actions',
//                'save_settings' => $base_button + [
//                    '#type' => 'submit',
//                    '#button_type' => 'primary',
//                    '#name' => $field_name . '_plugin_settings_update',
//                    '#value' => $this->t('Update'),
//                    '#op' => 'update',
//                  ],
//                'cancel_settings' => $base_button + [
//                    '#type' => 'submit',
//                    '#name' => $field_name . '_plugin_settings_cancel',
//                    '#value' => $this->t('Cancel'),
//                    '#op' => 'cancel',
//                    // Do not check errors for the 'Cancel' button, but make sure we
//                    // get the value of the 'plugin type' select.
//                    '#limit_validation_errors' => [['fields', $field_name, 'type']],
//                  ],
//              ],
//            ];
//            $field_row['#attributes']['class'][] = 'field-plugin-settings-editing';
//          }
//        }
//      }
//      else {
//        $field_row['settings_summary'] = [];
//        $field_row['settings_edit'] = [];
//
//        if ($plugin) {
//          // Display a summary of the current plugin settings, and (if the
//          // summary is not empty) a button to edit them.
//          $summary = $plugin->settingsSummary();
//
//          // Allow other modules to alter the summary.
//          $this->alterSettingsSummary($summary, $plugin, $field_definition);
//
//          if (!empty($summary)) {
//            $field_row['settings_summary'] = [
//              '#type' => 'inline_template',
//              '#template' => '<div class="field-plugin-summary">{{ summary|safe_join("<br />") }}</div>',
//              '#context' => ['summary' => $summary],
//              '#cell_attributes' => ['class' => ['field-plugin-summary-cell']],
//            ];
//          }
//
//          // Check selected plugin settings to display edit link or not.
//          $settings_form = $plugin->settingsForm($form, $form_state);
//          $third_party_settings_form = $this->thirdPartySettingsForm($plugin, $field_definition, $form, $form_state);
//          if (!empty($settings_form) || !empty($third_party_settings_form)) {
//            $field_row['settings_edit'] = $base_button + [
//                '#type' => 'image_button',
//                '#name' => $field_name . '_settings_edit',
//                '#src' => 'core/misc/icons/787878/cog.svg',
//                '#attributes' => ['class' => ['field-plugin-settings-edit'], 'alt' => $this->t('Edit')],
//                '#op' => 'edit',
//                // Do not check errors for the 'Edit' button, but make sure we get
//                // the value of the 'plugin type' select.
//                '#limit_validation_errors' => [['fields', $field_name, 'type']],
//                '#prefix' => '<div class="field-plugin-settings-edit-wrapper">',
//                '#suffix' => '</div>',
//              ];
//          }
//        }
//      }
//      $table[$fieldName]['harmonize_settings_summary'] = $table[$fieldName]['settings_summary'];
//      $table[$fieldName]['harmonize_settings_edit'] = $table[$fieldName]['settings_edit'];
    }

    // Attach our submit handler.
    $form['actions']['submit']['#submit'][] = get_class() . '::submitHandler';
  }

  /**
   * Custom submit handler.
   *
   * @noinspection PhpUnused
   * @noinspection PhpUnusedParameterInspection
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public static function submitHandler(array $form, FormStateInterface $formState) {
    // Get entity type ID and bundle.
    /* @noinspection PhpPossiblePolymorphicInvocationInspection */
    $formEntity = $formState->getFormObject()->getEntity();
    $entityTypeId = $formEntity->getTargetEntityTypeId();
    $bundle = $formEntity->getTargetBundle();

    // Load our settings entity.
    $settings = FormTuner::loadEntitySettings($entityTypeId, $bundle);

    // Defensive checks. Might be overkill but whatever.
    if (empty($settings)) {
      return;
    }

    // Load display settings.
    $displaySettings = $settings->getDisplaySettings();

    // Get the display mode.
    $mode = $formEntity->getMode();

    // Get our harmonize values.
    $fields = $formState->getValue('fields');

    // Loop through our fields and set fields.
    foreach ($fields as $fieldName => $fieldValues) {
      if (isset($fieldValues['harmonize'])) {
        $displaySettings[$mode][$fieldName] = (bool) $fieldValues['harmonize'];
      }
    }

    // Set the new display settings and save.
    $settings->setDisplaySettings($displaySettings);
    $settings->save();
  }

}
